#!/usr/bin/env python

"""
usage: flaskrest_test05.py [-h] [-v] [--host HOST] [--port PORT]

synopsis:
  Run flask-rest app with in-memory data store (a Python dict).
  Implements REST CRUD access to a set of documents (name, description).
  Uses content addressable data store.  Look-up id is generated from
    data with an SHA-3 hash function.

  Use function generate_hash in this module to generate hash/id.

optional arguments:
  -h, --help     show this help message and exit
  -v, --verbose  Print messages during actions.
  --host HOST    Host address from which to run the server. Default: 0.0.0.0
  --port PORT    Port from which to run the server. Default: 8001

supports the following request URLs:
  # Get a list of documents.
  '/documents/getdocument',
  # Get one document by name.
  '/documents/getdocument/<document_id>',
  # Add a document.  Form data: description.
  '/documents/adddocument',
  # Update a document.  Form data: document_id, description.
  '/documents/updatedocument',
  # Delete one document.`
  '/documents/delete/<string:document_id>',
"""

# Try to make our app run under both Python 2 and Python 3.
from __future__ import print_function
import sys
import argparse
from flask import Flask, request
from flask_restful import Api, Resource
import sha3


app = Flask(__name__)
api = Api(app)

#
# The following can be used for configuration.
##app.config.update(
##    DEBUG=True,
##    #SERVER_NAME='localhost:8080',
##    #SERVER_NAME='localhost:5000',
##    SERVER_NAME='localhost:8001',
##)

# Define the host and post.
# For access from this machine only:
#SERVER_HOST = 'localhost'
# For access across the network:
SERVER_HOST = '0.0.0.0'
SERVER_PORT = 8001

CMD_LINE_OPTIONS = None

DATA_STORE = {}


#
# Public functions:

def generate_hash(body):
    """Generate a 32-byte SHA-3 hash from body."""
    if sys.version_info.major == 3 and hasattr(body, 'encode'):
        body = body.encode('utf-8')
    hash = sha3.sha3_256(body).hexdigest()
    return hash


#
# Private functions:

def dbg_msg(msg):
    """Print a message if verbose is on."""
    if CMD_LINE_OPTIONS.verbose:
        print(msg)


class DocumentList(Resource):
    """Handle operations on the list of documents.

    Enable a GET command to get a list of documents or a single document.
    Enable a POST command to create and add a new document.
    Enable a PUT command to update a document.
    """
    def __init__(self, *args, **kwargs):
        super(DocumentList, self).__init__(*args, **kwargs)

    def get(self, document_id=None):
        """Return a collection of existing documents or a single document."""
        #
        # begin fill-in
        # ---------------------------------------
        if document_id is None:
            dbg_msg('Retrieving list of documents')
            collection = DATA_STORE
            content = {
                'action': 'get-list',
                'collection': collection}
        else:
            dbg_msg('Retrieving document: {}'.format(document_id))
            document = DATA_STORE.get(document_id)
            if document is not None:
                content = {
                    'action': 'get-item',
                    'document_id': document_id,
                    'body': document['body'],
                }
            else:
                content = {
                    'action': 'error',
                    'reason': 'document not found: {}'.format(document_id),
                }
        return content
        # ---------------------------------------
        # end fill-in
        #

    def post(self):
        """Create and save a new document."""
        #
        # begin fill-in
        # ---------------------------------------
        body = request.form['document']
        hash = generate_hash(body)
        document = DATA_STORE.get(hash)
        if document is not None:
            document_json = {
                'action': 'error',
                'reason': 'document {} already exists'.format(hash)}
        else:
            new_document = {
                'id': hash,
                'body': body,
            }
            DATA_STORE[hash] = new_document
            dbg_msg('Created document --\n    id: {}\n    body: {}'.format(
                hash, body))
            document_json = {
                'action': 'create',
                'id': hash,
                'body': body,
            }
        return document_json
        # ---------------------------------------
        # end fill-in
        #

    def put(self, document_id):
        """Update an existing document."""
        #
        # begin fill-in
        # ---------------------------------------
        old_hash = document_id
        body = request.form['document']
        new_hash = generate_hash(body)
        old_document = DATA_STORE.get(old_hash)
        if old_document is None:
            document_json = {
                'action': 'error',
                'reason': 'old document {} does not exist.'.format(old_hash)}
        else:
            del DATA_STORE[old_hash]
            new_document = {
                'id': new_hash,
                'body': body,
            }
            DATA_STORE[new_hash] = new_document
            dbg_msg('Created document --\n    id: {}\n    body: {}'.format(
                new_hash, body))
            document_json = {
                'action': 'update',
                'id': new_hash,
                'body': body,
            }
        return document_json
        # ---------------------------------------
        # end fill-in
        #

    def delete(self, document_id):
        """Delete a single document."""
        #
        # begin fill-in
        # ---------------------------------------
        hash = document_id
        dbg_msg('Deleting document: {}'.format(hash))
        document = DATA_STORE.get(hash)
        if document is None:
            document_json = {
                'action': 'error',
                'reason': 'document {} does not exist'.format(hash)}
        else:
            del DATA_STORE[hash]
            document_json = {
                'action': 'delete',
                'id': hash,
                'body': document,
            }
        return document_json
        #return '', 204
        # ---------------------------------------
        # end fill-in
        #


#
# Set up the api resource routing.
#
api.add_resource(
    DocumentList,
    # Get a list of documents.
    '/documents/getdocument',
    # Get one document by name.
    '/documents/getdocument/<string:document_id>',
    # Add a document.  Form data: body.
    '/documents/adddocument',
    # Update a document.  Form data: body.
    '/documents/updatedocument/<string:document_id>',
    # Delete one document.`
    '/documents/delete/<string:document_id>',
)


def main():
    global CMD_LINE_OPTIONS
    description = """\
synopsis:
  Run flask-rest app with in-memory data store (a Python dict).
  Implements REST CRUD access to a set of documents (name, description).
  Uses content addressable data store.  Look-up id is a 32-byte
    secure hash generated from data with an SHA-3 hash function.
  Use function generate_hash in this module to generate hash/id.
"""
    epilog = """\
supports the following request URLs:
  # Get a list of documents.
  '/documents/getdocument',
  # Get one document by name.
  '/documents/getdocument/<document_id>',
  # Add a document.  Form data: description.
  '/documents/adddocument',
  # Update a document.  Form data: document_id, description.
  '/documents/updatedocument',
  # Delete one document.`
  '/documents/delete/<string:document_id>',
"""
    parser = argparse.ArgumentParser(
        description=description,
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-v", "--verbose",
        action="store_true",
        help="Print messages during actions.",
    )
    parser.add_argument(
        "--host",
        help="Host address from which to run the server.  Default: {}".format(
            SERVER_HOST),
    )
    parser.add_argument(
        "--port",
        help="Port from which to run the server.  Default: {}".format(
            SERVER_PORT),
    )
    CMD_LINE_OPTIONS = parser.parse_args()
    host = CMD_LINE_OPTIONS.host if CMD_LINE_OPTIONS.host else SERVER_HOST
    port = int(CMD_LINE_OPTIONS.port) if CMD_LINE_OPTIONS.port else SERVER_PORT
    app.run(debug=True, host=host, port=port)


if __name__ == '__main__':
    #import ipdb; ipdb.set_trace()
    main()
